import logging
from django.db.models.query import QuerySet
from django.db import models
from fcm.utils import FCMMessage

logger = logging.getLogger(__name__)


class DeviceQuerySet(QuerySet):

    def send_message(self, data, **kwargs):
        if self:
            registration_ids = list(self.values_list("reg_id", flat=True))
            return FCMMessage().send(
                data, registration_ids=registration_ids, **kwargs)


class DeviceManager(models.Manager):

    def get_queryset(self):
        return DeviceQuerySet(self.model)


class AbstractDevice(models.Model):
    dev_id = models.CharField(verbose_name=("Device ID"), max_length=50, unique=True,)
    reg_id = models.CharField(verbose_name=("Registration ID"), max_length=255, unique=True)
    name = models.CharField(verbose_name=("Name"), max_length=255, blank=True, null=True)
    is_active = models.BooleanField(verbose_name=("Is active?"), default=False)

    objects = DeviceManager()

    def __str__(self):
        return self.dev_id

    class Meta:
        abstract = True
        verbose_name = ("Device")
        verbose_name_plural = ("Devices")

    def send_message(self, data, **kwargs):
        return FCMMessage().send(
            registration_ids=[self.reg_id], data=data, **kwargs)

    def mark_inactive(self, **kwargs):
        self.is_active = False
        self.save()
        if kwargs.get('error_message'):
            logger.debug("Device %s (%s) marked inactive due to error: %s",
                         self.dev_id, self.name, kwargs['error_message'])


class Device(AbstractDevice):
    class Meta:
        swappable = 'FCM_DEVICE_MODEL'



from django.conf.urls import url, include

from rest_framework import routers
from fcm.views import DeviceViewSet

router = routers.DefaultRouter()
router.register(r'devices', DeviceViewSet)

urlpatterns = [
    url(r'v1/', include(router.urls))

]



from django.core.management.base import BaseCommand
from django.core.urlresolvers import reverse
from fcm.views import DeviceViewSet


class Command(BaseCommand):
    help = "Show FCM urls"

    def show_line(self):
        self.stdout.write("%s\n" % ("-" * 30))
        
    def handle(self, **options):
        url_kwargs = {'api_name': 'v1'}
        # TODO add proper url resolvers
        #register_url = reverse('devices', kwargs=url_kwargs)
        #unregister_url = reverse('devices', kwargs=url_kwargs)
        register_url = 'fcm/v1/devices'
        unregister_url = 'fcm/v1/devices/<pk>'
        self.show_line()
        self.stdout.write("FCM urls:\n")
        self.stdout.write("* Register device\n    %s\n" % register_url)
        self.stdout.write("* Unregister device\n    %s\n" % unregister_url)
        self.show_line()



from rest_framework import viewsets, status, response
from fcm.utils import get_device_model
from fcm.serializers import DeviceSerializer
Device = get_device_model()


class DeviceViewSet(viewsets.ModelViewSet):
    queryset = Device.objects.all()
    serializer_class = DeviceSerializer

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=False)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return response.Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def perform_create(self, serializer):
        try:
            device = Device.objects.get(dev_id=serializer.data["dev_id"])
        except Device.DoesNotExist:
            device = Device(dev_id=serializer.data["dev_id"])
        device.is_active = True
        device.reg_id = serializer.data["reg_id"]
        device.name = serializer.data["name"]
        device.save()

    def destroy(self, request, *args, **kwargs):
        try:
            instance = Device.objects.get(dev_id=kwargs["pk"])
            self.perform_destroy(instance)
            return response.Response(status=status.HTTP_200_OK)
        except Device.DoesNotExist:
            return response.Response(status=status.HTTP_404_NOT_FOUND)

    def perform_destroy(self, instance):
        instance.is_active = False
        instance.save()



from optparse import make_option
from django.core.management.base import BaseCommand, CommandError

from fcm.utils import get_device_model

Device = get_device_model()


class Command(BaseCommand):
    args = ['<device_id>', '<message>']
    help = 'Send message through fcm api'

    def add_arguments(self, parser):
        #this store list
        parser.add_argument('--device_id', nargs='*', type=str)
        parser.add_argument('--msg', nargs='*', type=str)

        parser.add_argument(
            '--devices',
            action='store_true',
            dest='devices',
            default=False,
            help='List of available devices',
        )
        parser.add_argument(
            '--collapse-key',
            dest='collapse_key',
            default='message',
            help='Set value of collapse_key flag, default is "message"',
        )

    def handle(self, *args, **options):
        if options['devices']:
            devices = Device.objects.filter(is_active=True)

            self.stdout.write("Devices list:\n")
            for device in devices:
                self.stdout.write("(#%s) %s\n" % (device.id, device.name))
            self.stdout.write("\n")
        else:
            collapse_key = options['collapse_key']
            try:
                id = options['device_id'][0]
                message = options['msg'][0]
            except IndexError:
                raise CommandError(
                    "Invalid params. You have to put all params: "
                    "python manage.py fcm_messenger <device_id> <msg>")

            try:
                device = Device.objects.get(pk=int(id), is_active=True)
            except Device.DoesNotExist:
                raise CommandError(
                    'Unknown device (id=%s). Check list: '
                    'python manage.py fcm_messenger --devices' % id)
            else:
                result = device.send_message(
                        {'message': message}, collapse_key=collapse_key)

                self.stdout.write("[OK] device #%s (%s): %s\n" %
                                  (id, device.name, result))



from functools import update_wrapper
from django.contrib import admin
from django.shortcuts import redirect, render
from django.template import RequestContext
from django.utils.translation import ugettext_lazy as _

from .forms import MessageForm
from .models import get_device_model

Device = get_device_model()


class DeviceAdmin(admin.ModelAdmin):
    list_display = ['dev_id', 'name', 'modified_date', 'is_active']
    search_fields = ('dev_id', 'name')
    list_filter = ['is_active']
    date_hierarchy = 'modified_date'
    readonly_fields = ('dev_id', 'reg_id')
    actions = ['send_message_action']

    def get_urls(self):
        from django.conf.urls import url

        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(view)(*args, **kwargs)
            return update_wrapper(wrapper, view)

        urlpatterns = [
            url(r'^send-message/$', wrap(self.send_message_view),
                name=self.build_admin_url('send_message'))]
        return urlpatterns + super(DeviceAdmin, self).get_urls()

    def build_admin_url(self, url_name):
        return '%s_%s_%s' % (self.model._meta.app_label,
                             self.model._meta.model_name,
                             url_name)

    def send_message_view(self, request):
        base_view = 'admin:%s' % self.build_admin_url('changelist')
        session_key = 'device_ids'
        device_ids = request.session.get(session_key)
        if not device_ids:
            return redirect(base_view)

        form = MessageForm(data=request.POST or None)
        if form.is_valid():
            devices = Device.objects.filter(pk__in=device_ids)
            for device in devices:
                device.send_message(form.cleaned_data['message'])
            self.message_user(request, _('Message was sent.'))
            del request.session[session_key]
            return redirect(base_view)

        context = {'form': form, 'opts': self.model._meta, 'add': False}
        return render(request, 'fcm/admin/send_message.html', context)

    def send_message_action(self, request, queryset):
        ids = queryset.values_list('id', flat=True)
        request.session['device_ids'] = list(ids)
        url = 'admin:%s' % self.build_admin_url('send_message')
        return redirect(url)
    send_message_action.short_description = _("Send message")


admin.site.register(Device, DeviceAdmin)


import json
import requests
from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.utils.encoding import force_text
from django.conf import settings
from django.utils.module_loading import import_module

DEFAULT_FCM_DEVICE_MODEL = 'fcm.Device'


def get_device_model():
    model = getattr(settings, "FCM_DEVICE_MODEL", DEFAULT_FCM_DEVICE_MODEL)
    try:
        return apps.get_model(model)
    except ValueError:
        raise ImproperlyConfigured("FCM_DEVICE_MODEL must be of the form 'app_label.model_name'")
    except LookupError:
        raise ImproperlyConfigured(
            "FCM_DEVICE_MODEL refers to model '%s' that has not been installed" % settings.FCM_DEVICE_MODEL
        )


class FCMMessage(object):

    def __init__(self):
        """
        you will not reach to test self.api_key if it is not set in settings...
        """
        try:
            self.api_key = settings.FCM_APIKEY
        except AttributeError:
            raise ImproperlyConfigured(
                "You haven't set the 'FCM_APIKEY' setting yet.")

        """
        accessing settings.FCM_MAX_RECIPIENTS if not set
        will crash the app, it can be set to 1 by default
        """
        try:
            self.max_recipients = settings.FCM_MAX_RECIPIENTS
        except AttributeError:
            # some kind of warning would be nice
            print("Using default settings.FCM_MAX_RECIPIENTS value 1. Change it via settings")
            self.max_recipients = 1

    def _chunks(self, items, limit):
        """
            Yield successive chunks from list \a items with a minimum size \a limit
        """
        for i in range(0, len(items), limit):
            yield items[i:i + limit]

    def send(self, data, registration_ids=None, **kwargs):
        """
        Send a FCM message for one or more devices, using json data
        registration_ids: A list with the devices which will be receiving a message
        data: The dict data which will be send
        Optional params e.g.:
            collapse_key: A string to group messages
        For more info see the following documentation:
        https://developer.android.com/google/fcm/server-ref.html#send-downstream
        """

        if not isinstance(data, dict):
            data = {'msg': data}

        registration_ids = registration_ids or []

        if len(registration_ids) >self.max_recipients:
            ret = []
            for chunk in self._chunks(
                    registration_ids, settings.FCM_MAX_RECIPIENTS):
                ret.append(self.send(data, registration_ids=chunk, **kwargs))
            return ret

        values = {
            'data': data,
            'collapse_key': 'message'}
        if registration_ids:
            values.update({'registration_ids': registration_ids})
        values.update(kwargs)

        values = json.dumps(values)

        headers = {
            'UserAgent': "FCM-Server",
            'Content-Type': 'application/json',
            'Authorization': 'key=' + self.api_key}

        response = requests.post(
            url="https://fcm.googleapis.com/fcm/send",
            data=values, headers=headers)

        response.raise_for_status()
        return registration_ids, json.loads(force_text(response.content))


class FCMMessage(FCMMessage):
    FCM_INVALID_ID_ERRORS = ['InvalidRegistration',
                             'NotRegistered',
                             'MismatchSenderId']

    def send(self, data, registration_ids=None, **kwargs):
        response = super(FCMMessage, self).send(
            data, registration_ids=registration_ids, **kwargs)
        chunks = [response] if not isinstance(response, list) else response
        for chunk in chunks:
            self.post_send(*chunk)
        return response

    def post_send(self, registration_ids, response):
        if response.get('failure'):
            invalid_messages = dict(filter(
                lambda x: x[1].get('error') in self.FCM_INVALID_ID_ERRORS,
                zip(registration_ids, response.get('results'))))

            regs = list(invalid_messages.keys())
            for device in get_device_model().objects.filter(reg_id__in=regs):
                device.mark_inactive(
                    error_message=invalid_messages[device.reg_id]['error'])



from django import forms
from django.utils.translation import ugettext_lazy as _

from .models import get_device_model


class RegisterDeviceForm(forms.ModelForm):

    class Meta:
        model = get_device_model()
        fields = ('dev_id', 'reg_id', 'name',)

    def save(self, commit=True):
        self.instance.is_active = True
        return super(RegisterDeviceForm, self).save(commit)


class UnregisterDeviceForm(forms.ModelForm):

    class Meta:
        model = get_device_model()
        fields = ('dev_id',)

    def clean(self):
        if not self.instance.pk:
            raise forms.ValidationError(
                "Device '%s' does not exist" % self.cleaned_data['dev_id'])
        return self.cleaned_data

    def save(self, commit=True):
        self.instance.mark_inactive()
        return super(UnregisterDeviceForm, self).save(commit)


class MessageForm(forms.Form):

    message = forms.CharField(label=_('Message'), required=True)


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Device',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dev_id', models.CharField(max_length=50, unique=True, verbose_name='Device ID')),
                ('reg_id', models.CharField(max_length=255, unique=True, verbose_name='Registration ID')),
                ('name', models.CharField(blank=True, max_length=255, null=True, verbose_name='Name')),
                ('is_active', models.BooleanField(default=False, verbose_name='Is active?')),
            ],
            options={
                'verbose_name_plural': 'Devices',
                'verbose_name': 'Device',
                'abstract': False,
            },
        ),
    ]



from rest_framework import serializers
from fcm.utils import get_device_model
Device = get_device_model()


class DeviceSerializer(serializers.ModelSerializer):
    class Meta:
        model = Device
        fields = ('dev_id', 'reg_id', 'name', 'is_active')
